﻿Imports System.IO
Imports System.Net
Imports System.Text
Imports System.Web
Imports System.Text.RegularExpressions

Friend Class PostData

#Region " Private Members "
    
    ''' <summary>
    ''' The value of the User-agent HTTP header. The default value is Nothing.
    ''' </summary>
    Private ReadOnly m_userAgent As String = "Mozilla/5.0 (Windows NT 5.1; rv:22.0) Gecko/20100101 Firefox/22.0"

    ''' <summary>
    ''' The number of milliseconds to wait before the request times out. The default value is 100,000 milliseconds (100 seconds).
    ''' </summary>
    Private ReadOnly m_timeoutSeconds As Integer = 30

    ''' <summary>
    ''' The URL of the web application.
    ''' </summary>
    Private ReadOnly m_forumUrl As String = "http://www.vbforums.com/"

    ''' <summary>
    ''' The request method to use to contact the Internet resource. The default value is GET.
    ''' </summary>
    Private ReadOnly m_formMethod As String = "Post"

    ''' <summary>
    ''' The form's security token.
    ''' </summary>
    ''' <remarks></remarks>
    Private m_securityToken As String = Nothing

    ''' <summary>
    ''' A CookieContainer that contains the cookies associated with this request.
    ''' </summary>
    Private m_cookies As CookieContainer
#End Region ' Private Members

#Region " Regular Expressions "
    Private ReadOnly m_loginSuccessPattern As String = "Thank you for logging in,"
    Private ReadOnly m_loginSuccessMatch As New Regex(m_loginSuccessPattern, RegexOptions.IgnoreCase Or RegexOptions.Compiled)

    Private ReadOnly m_loginUsernamePasswordErrorPattern As String = "You have entered an invalid username or password."
    Private ReadOnly m_loginUsernamePasswordErrorMatch As New Regex(m_loginUsernamePasswordErrorPattern, RegexOptions.IgnoreCase Or RegexOptions.Compiled)

    Private ReadOnly m_loginSecurityTokenPattern As String = "(?<=var SECURITYTOKEN = "").+(?="";)"
    Private ReadOnly m_loginSecurityTokenMatch As New Regex(m_loginSecurityTokenPattern, RegexOptions.IgnoreCase Or RegexOptions.Compiled)

    Private ReadOnly m_securityTokenPattern As String = "(?<=<input type=""hidden"" name=""securitytoken"" value="").+(?="" />)"
    Private ReadOnly m_securityTokenMatch As New Regex(m_securityTokenPattern, RegexOptions.IgnoreCase Or RegexOptions.Compiled)

    ' v3.02 update: Added from ident's code here, http://www.vbforums.com/showthread.php?767379
    Private ReadOnly m_failedLoginQuotaPattern As String = "You have used up your failed login quota!"
    Private ReadOnly m_failedLoginQuotaMatch As New Regex(m_failedLoginQuotaPattern, RegexOptions.IgnoreCase Or RegexOptions.Compiled)

    ' v3.02 update: not used
    'Private ReadOnly m_logoutHrefMissingPattern As String = "Your submission could not be processed because a security token was missing."
    'Private ReadOnly m_logoutHrefMissingMatch As New Regex(m_logoutHrefMissingPattern, RegexOptions.IgnoreCase Or RegexOptions.Compiled)

    ' v3.02 update: not used
    'Private ReadOnly m_logoutSuccessPattern As String = "All cookies cleared!"
    'Private ReadOnly m_logoutSuccessMatch As New Regex(m_logoutSuccessPattern, RegexOptions.IgnoreCase Or RegexOptions.Compiled)

#End Region ' Regular Expressions

#Region " Post Data Patterns "
    Private Shared ReadOnly m_loginPostData As String = "vb_login_username={0}&vb_login_password=&vb_login_password_hint=Password{1}&s=&securitytoken={2}&do=login&vb_login_md5password={3}&vb_login_md5password_utf={3}"
#End Region ' Post Data Patterns

#Region "Relative phpBB3 URLs"
    Private ReadOnly m_homeHref As String = "forum.php"
    Private ReadOnly m_loginHref As String = "login.php?do=login"
    Private ReadOnly m_logoutHref As String = "login.php?do=logout&logouthash={0}"
    Private ReadOnly m_controlPanelHref As String = "usercp.php"
#End Region

#Region " Events "
    ''' <summary>
    ''' Represents the method that will handle the <see cref="PostData.LastError">PostData</see> event of <see cref="LastError" />.
    ''' </summary>
    ''' <remarks >
    ''' Displays all relative post data patterns including authenticity token.
    ''' </remarks>
    Public Event LastError As EventHandler(Of LastErrorEventArghs)

    ''' <summary>
    ''' Represents the method that will handle the <see cref="PostData.ReturnedPageSource">PostData</see> event of <see cref="ReturnedPageSource" />.
    ''' </summary>
    ''' <remarks >
    ''' The most recent HTML source retrieved from the server.
    ''' </remarks>
    Public Event ReturnedPageSource As EventHandler(Of ReturnedPageSourceEventArghs)

    ''' <summary>
    ''' Represents the method that will handle the <see cref="PostData.ControlPanelSuccess">PostData</see> event of <see cref="ControlPanelSuccess" />.
    ''' </summary>
    ''' <remarks >
    ''' Control panel was successful.
    ''' </remarks>
    'Public Event ControlPanelSuccess As EventHandler '<<<< v3.02 update: not used

    ' v3.02 update: not used
    'Public Event LoginFailed()

#End Region ' Events

#Region " Overridable Events "
    Protected Overridable Sub OnLastError(ByVal e As LastErrorEventArghs)
        RaiseEvent LastError(Me, e)
    End Sub

    Protected Overridable Sub OnReturnedPageSource(ByVal e As ReturnedPageSourceEventArghs)
        RaiseEvent ReturnedPageSource(Me, e)
    End Sub

    ' v3.02 update: not used
    'Protected Overridable Sub OnControlPanelSuccess(ByVal e As EventArgs)
    '    RaiseEvent ControlPanelSuccess(Me, e)
    'End Sub

#End Region ' Overridable Events

    ''' <summary>
    ''' Attempts to login to the server with the supplied username and password.
    ''' First IsLoggedIn() is called, if this returns true, it is handled as an
    ''' error.
    ''' </summary>
    ''' <param name="username">Plaintext username</param>
    ''' <param name="password">MD5 Hash password</param>
    ''' <param name="cookieUser">If true, the session is remembered by the server.</param>
    ''' <remarks>
    ''' </remarks>
    Public Sub TryLogin(ByVal userName As String, ByVal password As String, ByVal cookieUser As Boolean)
        If Me.IsLoggedIn() Then
            Me.OnLastError(New LastErrorEventArghs(New Exception("User already logged in")))
            Exit Sub
        End If

        Me.m_cookies = New CookieContainer

        ' If user/pass fails it still downloads a web page and user is allowed in as a guest?
        If Me.m_securityToken = Nothing Then
            Me.m_securityToken = "guest"
        End If

        Dim loginUri As Uri = Me.GetAbsoluteUri(Me.m_loginHref)

        If loginUri Is Nothing Then
            Exit Sub
        End If

        Dim loginRequest As HttpWebRequest = TryCast(HttpWebRequest.Create(loginUri), HttpWebRequest)

        If loginRequest Is Nothing Then
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Failed to create HttpWebRequest for '{0}'.", loginUri.ToString()))))
            Exit Sub
        End If

        Dim loginPostDataFormatted As String = String.Format(m_loginPostData,
                                                             HttpUtility.HtmlEncode(userName),
                                                             If(cookieUser, "&cookieuser=1", String.Empty),
                                                             HttpUtility.HtmlEncode(Me.m_securityToken),
                                                             HttpUtility.HtmlEncode(password))

        Dim loginPostDataBuffer As Byte() = Encoding.ASCII.GetBytes(loginPostDataFormatted)

        loginRequest.UserAgent = Me.m_userAgent
        loginRequest.Timeout = Me.m_timeoutSeconds * 1000
        loginRequest.Method = "post"
        loginRequest.ContentType = "application/x-www-form-urlencoded"
        loginRequest.ContentLength = loginPostDataBuffer.Length
        loginRequest.CookieContainer = Me.m_cookies

        'if we don't have a web connection it will crash, Ed added a try-catch...
        Try
            Using requestStream As Stream = loginRequest.GetRequestStream() ' < crashed here if no connnection!
                requestStream.Write(loginPostDataBuffer, 0, loginPostDataBuffer.Length)
                requestStream.Close()
            End Using
        Catch ex As Exception
            ' v3.02 update: shortened error message.
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Failed to get response " & Environment.NewLine & "Check internet connection!"))))
            'Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Failed to get response from '{0}'." & Environment.NewLine & Environment.NewLine & "Make sure you have a web connection!" & Environment.NewLine, loginUri.ToString()), ex)))
            Exit Sub
        End Try

        Dim loginResponse As HttpWebResponse

        Try
            loginResponse = TryCast(loginRequest.GetResponse(), HttpWebResponse)
        Catch wex As WebException
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Request from '{0}' returned Status '{1}'.  See InnerException for more details.", loginUri.ToString(), wex.Status.ToString()), wex)))
            Exit Sub
        Catch ex As Exception
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Failed to get response from '{0}'.  See InnerException for more details.", loginUri.ToString()), ex)))
            Exit Sub
        End Try

        If loginResponse.StatusCode <> HttpStatusCode.OK Then
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Request from '{0}' returned StatusCode '{1}' ({2})", loginUri.ToString(), CInt(loginResponse.StatusCode), loginResponse.StatusCode.ToString()))))
        End If

        Dim loginSource As String = Nothing

        Using resposeStream As Stream = loginResponse.GetResponseStream()
            Using resposeReader As New StreamReader(resposeStream)
                loginSource = resposeReader.ReadToEnd()
            End Using
        End Using

        ' old ocde
        'If Not m_loginSuccessMatch.IsMatch(loginSource) Then
        '    If m_loginUsernamePasswordErrorMatch.IsMatch(loginSource) Then
        '        Me.OnLastError(New LastErrorEventArghs(New Exception(Me.m_loginUsernamePasswordErrorPattern)))
        '        Exit Sub ' << Ed added - problem if name/password were wrong.
        '    End If
        'End If
        ' new code(untested), v3.02 update: from ident's code here, http://www.vbforums.com/showthread.php?767379
        If Not Me.m_loginSuccessMatch.IsMatch(loginSource) Then
            Me.m_cookies = Nothing
            If Me.m_loginUsernamePasswordErrorMatch.IsMatch(loginSource) Then
                Me.OnLastError(New LastErrorEventArghs(New Exception(Me.m_loginUsernamePasswordErrorPattern)))
                Exit Sub
            ElseIf Me.m_failedLoginQuotaMatch.IsMatch(loginSource) Then
                Me.OnLastError(New LastErrorEventArghs(New Exception(Me.m_failedLoginQuotaPattern)))
                Exit Sub
            End If
        End If

        Me.m_cookies.Add(loginResponse.Cookies)
        NavigateUserControlPanel()
    End Sub

    ''' <summary>
    ''' Attempts to request to user control panel and returns the pages html.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub NavigateUserControlPanel()
        Dim controlUri As Uri = Me.GetAbsoluteUri(Me.m_controlPanelHref)

        If controlUri Is Nothing Then
            Exit Sub
        End If

        Dim controlRequest As HttpWebRequest = TryCast(HttpWebRequest.Create(controlUri), HttpWebRequest)

        If controlRequest Is Nothing Then
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Failed to create HttpWebRequest for '{0}'.", controlUri.ToString()))))
            Exit Sub
        End If

        controlRequest.UserAgent = Me.m_userAgent
        controlRequest.Timeout = Me.m_timeoutSeconds * 1000
        controlRequest.ContentType = "application/x-www-form-urlencoded"
        controlRequest.CookieContainer = Me.m_cookies

        Dim controlResponse As HttpWebResponse

        Try
            controlResponse = TryCast(controlRequest.GetResponse(), HttpWebResponse)
        Catch wex As WebException
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Request from '{0}' returned Status '{1}'.  See InnerException for more details.", controlUri.ToString(), wex.Status.ToString()), wex)))
            Exit Sub
        Catch ex As Exception
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Failed to get response from '{0}'.  See InnerException for more details.", controlUri.ToString()), ex)))
            Exit Sub
        End Try

        If controlResponse.StatusCode <> HttpStatusCode.OK Then
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Request from '{0}' returned StatusCode '{1}' ({2})", controlUri.ToString(), CInt(controlResponse.StatusCode), controlResponse.StatusCode.ToString()))))
        End If

        Dim controlSource As String

        ' Ed added "Encoding.GetEncoding("ISO-8859-1"), words like "Españolita" were returned as "Espa�olita".
        ' ident's original code didn't have an Encoding setting.
        ' see, http://www.vbforums.com/showthread.php?725347
        Using resposeStream As Stream = controlResponse.GetResponseStream()
            Using resposeReader As New StreamReader(resposeStream, Encoding.GetEncoding("ISO-8859-1"))
                controlSource = resposeReader.ReadToEnd()
            End Using
        End Using

        Me.m_cookies.Add(controlResponse.Cookies)
        Me.m_securityToken = HttpUtility.HtmlDecode(m_loginSecurityTokenMatch.Match(controlSource).Value)
        Me.OnReturnedPageSource(New ReturnedPageSourceEventArghs(controlSource))
    End Sub

    ''' <summary>
    ''' Forces a logout of the current session. Regardless of the return value, the session will no longer
    ''' be useable after this call. The use must be logged in again.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub TryLogout()
        If Not Me.IsLoggedIn() Then
            Me.OnLastError(New LastErrorEventArghs(New Exception("User not logged in")))
            Exit Sub
        End If

        Dim href As String = String.Format(Me.m_logoutHref, Me.m_securityToken)
        Dim logoutUri As Uri = Me.GetAbsoluteUri(href)

        If logoutUri Is Nothing Then
            Exit Sub
        End If

        Dim logoutRequest As HttpWebRequest = TryCast(HttpWebRequest.Create(logoutUri), HttpWebRequest)

        If logoutRequest Is Nothing Then
            Me.OnLastError(New LastErrorEventArghs(New Exception(String.Format("Failed to create HttpWebRequest for '{0}'.", logoutUri.ToString()))))
            Exit Sub
        End If

        logoutRequest.Timeout = Me.m_timeoutSeconds * 1000
        logoutRequest.CookieContainer = Me.m_cookies

        Dim logoutResponse As HttpWebResponse

        Try
            logoutResponse = TryCast(logoutRequest.GetResponse(), HttpWebResponse)
        Catch wex As WebException
            Me.OnLastError(New LastErrorEventArghs(New Exception(
                            String.Format("Request from '{0}' returned Status '{1}'.  See InnerException for more details.", logoutUri.ToString(), wex.Status.ToString()), wex)))
            Exit Sub
        Catch ex As Exception
            Me.OnLastError(New LastErrorEventArghs(New Exception(
                           String.Format("Failed to get response from '{0}'.  See InnerException for more details.", logoutUri.ToString()), ex)))
            Exit Sub
        End Try

        If logoutResponse.StatusCode <> HttpStatusCode.OK Then
            Me.OnLastError(New LastErrorEventArghs(New Exception(
                            String.Format("Request from '{0}' returned StatusCode '{1}' ({2})", logoutUri.ToString(), CInt(logoutResponse.StatusCode), logoutResponse.StatusCode.ToString()))))

        End If

        Dim logoutSource As String

        Using resposeStream As Stream = logoutResponse.GetResponseStream()
            Using resposeReader As New StreamReader(resposeStream)
                logoutSource = resposeReader.ReadToEnd()
            End Using
        End Using

        Me.m_cookies = Nothing
        Me.m_securityToken = Nothing

        Me.m_securityToken = HttpUtility.HtmlDecode(m_securityTokenMatch.Match(logoutSource).Value)
        Me.OnReturnedPageSource(New ReturnedPageSourceEventArghs(logoutSource))
    End Sub

#Region " Validation Functions "
    ''' <summary>
    ''' Get's an absolute URI based on the ForumURL and a path relative to it.
    ''' </summary>
    ''' <param name="href">The relative path.</param>
    ''' <returns>An absolute URI.</returns>
    ''' <remarks></remarks>
    Private Function GetAbsoluteUri(ByVal href As String) As Uri
        Dim result As Uri = Nothing

        If Not Uri.TryCreate(Me.m_forumUrl, UriKind.RelativeOrAbsolute, result) Then
            Me.OnLastError(New LastErrorEventArghs(New UriFormatException(
                                                   String.Format("The URI string '{0}' is invalid.", Me.m_forumUrl))))
            Return Nothing
        End If

        If Not Uri.TryCreate(result, href, result) Then
            Me.OnLastError(New LastErrorEventArghs(New UriFormatException(
                                                   String.Format("The relative URI string '{0}' is invalid.", href))))
            Return Nothing
        End If

        Return result
    End Function

    ''' <summary>
    ''' Returns True if this PostData object is associated with an authenticated session,
    ''' otherwise False.
    ''' </summary>
    ''' <returns>True is cookies not expired.</returns>
    ''' <remarks></remarks>
    Public Function IsLoggedIn() As Boolean

        If Me.m_cookies Is Nothing OrElse Me.m_cookies.Count = 0 Then
            Return False
        End If

        Dim cookieCollection As CookieCollection = Me.m_cookies.GetCookies(New Uri(Me.m_forumUrl))

        If cookieCollection.Count > 0 Then
            Dim expired As Boolean

            For Each cookie As Cookie In cookieCollection
                expired = expired OrElse cookie.Expired
            Next

            If Not expired Then
                Return True
            End If
        End If

        Return False
    End Function
#End Region ' Validation Functions

End Class